@@ -31,9 +31,9 @@ from .core import (
     bytes2int,
     Version,
     Tlv,
-    CommandError,
     NotSupportedError,
     BadResponseError,
+    InvalidPinError,
 )
 from .core.smartcard import (
     SW,
@@ -48,9 +48,15 @@ from cryptography import x509
 from cryptography.hazmat.primitives import hashes
 from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
 from cryptography.hazmat.primitives.constant_time import bytes_eq
-from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
-from cryptography.hazmat.primitives.asymmetric import rsa, ec
+from cryptography.hazmat.primitives.serialization import (
+    Encoding,
+    PublicFormat,
+    PrivateFormat,
+    NoEncryption,
+)
+from cryptography.hazmat.primitives.asymmetric import rsa, ec, ed25519, x25519
 from cryptography.hazmat.primitives.asymmetric.padding import AsymmetricPadding
+from cryptography.hazmat.primitives.asymmetric.utils import Prehashed
 from cryptography.hazmat.backends import default_backend
 
 from dataclasses import dataclass
@@ -58,6 +64,7 @@ from enum import Enum, IntEnum, unique
 from typing import Optional, Union, Type, cast
 
 import logging
+import gzip
 import os
 import re
 
@@ -82,15 +89,24 @@ def require_version(my_version: Version, *args, **kwargs):
 class KEY_TYPE(IntEnum):
     RSA1024 = 0x06
     RSA2048 = 0x07
+    RSA3072 = 0x05
+    RSA4096 = 0x16
     ECCP256 = 0x11
     ECCP384 = 0x14
+    ED25519 = 0xE0
+    X25519 = 0xE1
+
+    def __str__(self):
+        return self.name
 
     @property
     def algorithm(self):
-        return ALGORITHM.EC if self.name.startswith("ECC") else ALGORITHM.RSA
+        return ALGORITHM.RSA if self.name.startswith("RSA") else ALGORITHM.EC
 
     @property
     def bit_len(self):
+        if self in (KEY_TYPE.ED25519, KEY_TYPE.X25519):
+            return 256
         match = re.search(r"\d+$", self.name)
         if match:
             return int(match.group())
@@ -103,7 +119,6 @@ class KEY_TYPE(IntEnum):
                 return getattr(cls, "RSA%d" % key.key_size)
             except AttributeError:
                 raise ValueError("Unsupported RSA key size: %d" % key.key_size)
-                pass  # Fall through to ValueError
         elif isinstance(key, ec.EllipticCurvePublicKey):
             curve_name = key.curve.name
             if curve_name == "secp256r1":
@@ -111,6 +126,10 @@ class KEY_TYPE(IntEnum):
             elif curve_name == "secp384r1":
                 return cls.ECCP384
             raise ValueError(f"Unsupported EC curve: {curve_name}")
+        elif isinstance(key, ed25519.Ed25519PublicKey):
+            return cls.ED25519
+        elif isinstance(key, x25519.X25519PublicKey):
+            return cls.X25519
         raise ValueError(f"Unsupported key type: {type(key).__name__}")
 
 
@@ -142,8 +161,9 @@ def _parse_management_key(key_type, management_key):
         return algorithms.AES(management_key)
 
 
-# The card management slot is special, we don't include it in SLOT below
+# The following slots are special, we don't include it in SLOT below
 SLOT_CARD_MANAGEMENT = 0x9B
+SLOT_OCC_AUTH = 0x96
 
 
 @unique
@@ -176,6 +196,9 @@ class SLOT(IntEnum):
 
     ATTESTATION = 0xF9
 
+    def __str__(self) -> str:
+        return f"{int(self):02X} ({self.name})"
+
 
 @unique
 class OBJECT_ID(IntEnum):
@@ -227,6 +250,8 @@ class PIN_POLICY(IntEnum):
     NEVER = 0x1
     ONCE = 0x2
     ALWAYS = 0x3
+    MATCH_ONCE = 0x4
+    MATCH_ALWAYS = 0x5
 
 
 @unique
@@ -245,6 +270,7 @@ DEFAULT_MANAGEMENT_KEY = (
 )
 
 PIN_LEN = 8
+TEMPORARY_PIN_LEN = 16
 
 # Instruction set
 INS_VERIFY = 0x20
@@ -254,6 +280,7 @@ INS_GENERATE_ASYMMETRIC = 0x47
 INS_AUTHENTICATE = 0x87
 INS_GET_DATA = 0xCB
 INS_PUT_DATA = 0xDB
+INS_MOVE_KEY = 0xF6
 INS_GET_METADATA = 0xF7
 INS_ATTEST = 0xF9
 INS_SET_PIN_RETRIES = 0xFA
@@ -284,6 +311,8 @@ TAG_METADATA_ORIGIN = 0x03
 TAG_METADATA_PUBLIC_KEY = 0x04
 TAG_METADATA_IS_DEFAULT = 0x05
 TAG_METADATA_RETRIES = 0x06
+TAG_METADATA_BIO_CONFIGURED = 0x07
+TAG_METADATA_TEMPORARY_PIN = 0x08
 
 ORIGIN_GENERATED = 1
 ORIGIN_IMPORTED = 2
@@ -295,14 +324,7 @@ INDEX_RETRIES_REMAINING = 1
 
 PIN_P2 = 0x80
 PUK_P2 = 0x81
-
-
-class InvalidPinError(CommandError):
-    def __init__(self, attempts_remaining):
-        super(InvalidPinError, self).__init__(
-            "Invalid PIN/PUK. Remaining attempts: %d" % attempts_remaining
-        )
-        self.attempts_remaining = attempts_remaining
+UV_P2 = 0x96
 
 
 def _pin_bytes(pin):
@@ -312,15 +334,13 @@ def _pin_bytes(pin):
     return pin.ljust(PIN_LEN, b"\xff")
 
 
-def _retries_from_sw(version, sw):
+def _retries_from_sw(sw):
     if sw == SW.AUTH_METHOD_BLOCKED:
         return 0
-    if version < (1, 0, 4):
-        if 0x6300 <= sw <= 0x63FF:
-            return sw & 0xFF
-    else:
-        if 0x63C0 <= sw <= 0x63CF:
-            return sw & 0x0F
+    if sw & 0xFFF0 == 0x63C0:
+        return sw & 0x0F
+    elif sw & 0xFF00 == 0x6300:
+        return sw & 0xFF
     return None
 
 
@@ -351,11 +371,23 @@ class SlotMetadata:
         return _parse_device_public_key(self.key_type, self.public_key_encoded)
 
 
+@dataclass
+class BioMetadata:
+    configured: bool
+    attempts_remaining: int
+    temporary_pin: bool
+
+
 def _pad_message(key_type, message, hash_algorithm, padding):
+    if key_type in (KEY_TYPE.ED25519, KEY_TYPE.X25519):
+        return message
     if key_type.algorithm == ALGORITHM.EC:
-        h = hashes.Hash(hash_algorithm, default_backend())
-        h.update(message)
-        hashed = h.finalize()
+        if isinstance(hash_algorithm, Prehashed):
+            hashed = message
+        else:
+            h = hashes.Hash(hash_algorithm, default_backend())
+            h.update(message)
+            hashed = h.finalize()
         byte_len = key_type.bit_len // 8
         if len(hashed) < byte_len:
             return hashed.rjust(byte_len // 8, b"\0")
@@ -415,6 +447,21 @@ def check_key_support(
         if pin_policy == PIN_POLICY.NEVER:
             raise NotSupportedError("PIN_POLICY.NEVER not allowed on YubiKey FIPS")
 
+    # New key types
+    if version < (5, 7, 0) and key_type in (
+        KEY_TYPE.RSA3072,
+        KEY_TYPE.RSA4096,
+        KEY_TYPE.ED25519,
+        KEY_TYPE.X25519,
+    ):
+        raise NotSupportedError(f"{key_type} requires YubiKey 5.7 or later")
+
+    # TODO: Detect Bio capabilities
+    if version < () and pin_policy in (PIN_POLICY.MATCH_ONCE, PIN_POLICY.MATCH_ALWAYS):
+        raise NotSupportedError(
+            "Biometric match PIN policy requires YubiKey 5.6 or later"
+        )
+
 
 def _parse_device_public_key(key_type, encoded):
     data = Tlv.parse_dict(encoded)
@@ -422,6 +469,10 @@ def _parse_device_public_key(key_type, encoded):
         modulus = bytes2int(data[0x81])
         exponent = bytes2int(data[0x82])
         return rsa.RSAPublicNumbers(exponent, modulus).public_key(default_backend())
+    elif key_type == KEY_TYPE.ED25519:
+        return ed25519.Ed25519PublicKey.from_public_bytes(data[0x86])
+    elif key_type == KEY_TYPE.X25519:
+        return x25519.X25519PublicKey.from_public_bytes(data[0x86])
     else:
         if key_type == KEY_TYPE.ECCP256:
             curve: Type[ec.EllipticCurve] = ec.SECP256R1
@@ -432,6 +483,8 @@ def _parse_device_public_key(key_type, encoded):
 
 
 class PivSession:
+    """A session with the PIV application."""
+
     def __init__(self, connection: SmartCardConnection):
         self.protocol = SmartCardProtocol(connection)
         self.protocol.select(AID.PIV)
@@ -452,6 +505,14 @@ class PivSession:
     def reset(self) -> None:
         logger.debug("Preparing PIV reset")
 
+        try:
+            if self.get_bio_metadata().configured:
+                raise ValueError(
+                    "Cannot perform PIV reset when biometrics are configured"
+                )
+        except NotSupportedError:
+            pass
+
         # Block PIN
         logger.debug("Verify PIN with invalid attempts until blocked")
         counter = self.get_pin_attempts()
@@ -464,7 +525,10 @@ class PivSession:
 
         # Block PUK
         logger.debug("Verify PUK with invalid attempts until blocked")
-        counter = 1
+        try:
+            counter = self.get_puk_metadata().attempts_remaining
+        except NotSupportedError:
+            counter = 1
         while counter > 0:
             try:
                 self._change_reference(INS_RESET_RETRY, PIN_P2, "", "")
@@ -483,6 +547,11 @@ class PivSession:
     def authenticate(
         self, key_type: MANAGEMENT_KEY_TYPE, management_key: bytes
     ) -> None:
+        """Authenticate to PIV with management key.
+
+        :param key_type: The management key type.
+        :param management_key: The management key in raw bytes.
+        """
         key_type = MANAGEMENT_KEY_TYPE(key_type)
         logger.debug(f"Authenticating with key type: {key_type}")
         response = self.protocol.send_apdu(
@@ -523,6 +592,12 @@ class PivSession:
         management_key: bytes,
         require_touch: bool = False,
     ) -> None:
+        """Set a new management key.
+
+        :param key_type: The management key type.
+        :param management_key: The management key in raw bytes.
+        :param require_touch: The touch policy.
+        """
         key_type = MANAGEMENT_KEY_TYPE(key_type)
         logger.debug(f"Setting management key of type: {key_type}")
 
@@ -541,18 +616,57 @@ class PivSession:
         logger.info("Management key set")
 
     def verify_pin(self, pin: str) -> None:
+        """Verify the PIN.
+
+        :param pin: The PIN.
+        """
         logger.debug("Verifying PIN")
         try:
             self.protocol.send_apdu(0, INS_VERIFY, 0, PIN_P2, _pin_bytes(pin))
             self._current_pin_retries = self._max_pin_retries
         except ApduError as e:
-            retries = _retries_from_sw(self.version, e.sw)
+            retries = _retries_from_sw(e.sw)
             if retries is None:
                 raise
             self._current_pin_retries = retries
             raise InvalidPinError(retries)
 
+    def verify_uv(self) -> bytes:
+        logger.debug("Verifying UV")
+        try:
+            return self.protocol.send_apdu(0, INS_VERIFY, 0, SLOT_OCC_AUTH)
+        except ApduError as e:
+            if e.sw == SW.REFERENCE_DATA_NOT_FOUND:
+                raise NotSupportedError(
+                    "Biometric verification not supported by this YuibKey"
+                )
+            retries = _retries_from_sw(e.sw)
+            if retries is None:
+                raise
+            raise InvalidPinError(
+                retries, f"Fingerprint mismatch, {retries} attempts remaining"
+            )
+
+    def verify_temporary_pin(self, pin: bytes) -> None:
+        logger.debug("Verifying temporary PIN")
+        if len(pin) != TEMPORARY_PIN_LEN:
+            raise ValueError(f"Temporary PIN must be exactly {TEMPORARY_PIN_LEN} bytes")
+        try:
+            self.protocol.send_apdu(0, INS_VERIFY, 0, SLOT_OCC_AUTH, Tlv(1, pin))
+        except ApduError as e:
+            if e.sw == SW.REFERENCE_DATA_NOT_FOUND:
+                raise NotSupportedError(
+                    "Biometric verification not supported by this YuibKey"
+                )
+            retries = _retries_from_sw(e.sw)
+            if retries is None:
+                raise
+            raise InvalidPinError(
+                retries, f"Invalid temporary PIN, {retries} attempts remaining"
+            )
+
     def get_pin_attempts(self) -> int:
+        """Get remaining PIN attempts."""
         logger.debug("Getting PIN attempts")
         try:
             return self.get_pin_metadata().attempts_remaining
@@ -563,7 +677,7 @@ class PivSession:
                 logger.debug("Using cached value, may be incorrect.")
                 return self._current_pin_retries
             except ApduError as e:
-                retries = _retries_from_sw(self.version, e.sw)
+                retries = _retries_from_sw(e.sw)
                 if retries is None:
                     raise
                 self._current_pin_retries = retries
@@ -571,36 +685,70 @@ class PivSession:
                 return retries
 
     def change_pin(self, old_pin: str, new_pin: str) -> None:
+        """Change the PIN.
+
+        :param old_pin: The current PIN.
+        :param new_pin: The new PIN.
+        """
         logger.debug("Changing PIN")
         self._change_reference(INS_CHANGE_REFERENCE, PIN_P2, old_pin, new_pin)
         logger.info("New PIN set")
 
     def change_puk(self, old_puk: str, new_puk: str) -> None:
+        """Change the PUK.
+
+        :param old_puk: The current PUK.
+        :param new_puk: The new PUK.
+        """
         logger.debug("Changing PUK")
         self._change_reference(INS_CHANGE_REFERENCE, PUK_P2, old_puk, new_puk)
         logger.info("New PUK set")
 
     def unblock_pin(self, puk: str, new_pin: str) -> None:
+        """Reset PIN with PUK.
+
+        :param puk: The PUK.
+        :param new_pin: The new PIN.
+        """
         logger.debug("Using PUK to set new PIN")
         self._change_reference(INS_RESET_RETRY, PIN_P2, puk, new_pin)
         logger.info("New PIN set")
 
     def set_pin_attempts(self, pin_attempts: int, puk_attempts: int) -> None:
+        """Set PIN retries for PIN and PUK.
+
+        Both PIN and PUK will be reset to default values when this is executed.
+
+        Requires authentication with management key and PIN verification.
+
+        :param pin_attempts: The PIN attempts.
+        :param puk_attempts: The PUK attempts.
+        """
         logger.debug(f"Setting PIN/PUK attempts ({pin_attempts}, {puk_attempts})")
-        self.protocol.send_apdu(0, INS_SET_PIN_RETRIES, pin_attempts, puk_attempts)
-        self._max_pin_retries = pin_attempts
-        self._current_pin_retries = pin_attempts
-        logger.info("PIN/PUK attempts set")
+        try:
+            self.protocol.send_apdu(0, INS_SET_PIN_RETRIES, pin_attempts, puk_attempts)
+            self._max_pin_retries = pin_attempts
+            self._current_pin_retries = pin_attempts
+            logger.info("PIN/PUK attempts set")
+        except ApduError as e:
+            if e.sw == SW.INVALID_INSTRUCTION:
+                raise NotSupportedError(
+                    "Setting PIN attempts not supported on this YubiKey"
+                )
+            raise
 
     def get_pin_metadata(self) -> PinMetadata:
+        """Get PIN metadata."""
         logger.debug("Getting PIN metadata")
         return self._get_pin_puk_metadata(PIN_P2)
 
     def get_puk_metadata(self) -> PinMetadata:
+        """Get PUK metadata."""
         logger.debug("Getting PUK metadata")
         return self._get_pin_puk_metadata(PUK_P2)
 
     def get_management_key_metadata(self) -> ManagementKeyMetadata:
+        """Get management key metadata."""
         logger.debug("Getting management key metadata")
         require_version(self.version, (5, 3, 0))
         data = Tlv.parse_dict(
@@ -614,6 +762,10 @@ class PivSession:
         )
 
     def get_slot_metadata(self, slot: SLOT) -> SlotMetadata:
+        """Get slot metadata.
+
+        :param slot: The slot to get metadata from.
+        """
         slot = SLOT(slot)
         logger.debug(f"Getting metadata for slot {slot}")
         require_version(self.version, (5, 3, 0))
@@ -627,6 +779,24 @@ class PivSession:
             data[TAG_METADATA_PUBLIC_KEY],
         )
 
+    def get_bio_metadata(self) -> BioMetadata:
+        logger.debug("Getting bio metadata")
+        try:
+            data = Tlv.parse_dict(
+                self.protocol.send_apdu(0, INS_GET_METADATA, 0, SLOT_OCC_AUTH)
+            )
+        except ApduError as e:
+            if e.sw in (SW.REFERENCE_DATA_NOT_FOUND, SW.INVALID_INSTRUCTION):
+                raise NotSupportedError(
+                    "Biometric verification not supported by this YuibKey"
+                )
+            raise
+        return BioMetadata(
+            1 == data.get(TAG_METADATA_BIO_CONFIGURED, b"\x00")[0],
+            data[TAG_METADATA_RETRIES][0],
+            1 == data.get(TAG_METADATA_TEMPORARY_PIN, b"\x00")[0],
+        )
+
     def sign(
         self,
         slot: SLOT,
@@ -635,6 +805,16 @@ class PivSession:
         hash_algorithm: hashes.HashAlgorithm,
         padding: Optional[AsymmetricPadding] = None,
     ) -> bytes:
+        """Sign message with key.
+
+        Requires PIN verification.
+
+        :param slot: The slot of the key to use.
+        :param key_type: The type of the key to sign with.
+        :param message: The message to sign.
+        :param hash_algorithm: The pre-signature hash algorithm to use.
+        :param padding: The pre-signature padding.
+        """
         slot = SLOT(slot)
         key_type = KEY_TYPE(key_type)
         logger.debug(
@@ -647,12 +827,18 @@ class PivSession:
     def decrypt(
         self, slot: SLOT, cipher_text: bytes, padding: AsymmetricPadding
     ) -> bytes:
+        """Decrypt cipher text.
+
+        Requires PIN verification.
+
+        :param slot: The slot.
+        :param cipher_text: The cipher text to decrypt.
+        :param padding: The padding of the plain text.
+        """
         slot = SLOT(slot)
-        if len(cipher_text) == 1024 // 8:
-            key_type = KEY_TYPE.RSA1024
-        elif len(cipher_text) == 2048 // 8:
-            key_type = KEY_TYPE.RSA2048
-        else:
+        try:
+            key_type = getattr(KEY_TYPE, f"RSA{len(cipher_text) * 8}")
+        except AttributeError:
             raise ValueError("Invalid length of ciphertext")
         logger.debug(
             f"Decrypting data with key in slot {slot} of type {key_type} using ",
@@ -662,8 +848,19 @@ class PivSession:
         return _unpad_message(padded, padding)
 
     def calculate_secret(
-        self, slot: SLOT, peer_public_key: ec.EllipticCurvePublicKey
+        self,
+        slot: SLOT,
+        peer_public_key: Union[
+            ec.EllipticCurvePrivateKeyWithSerialization, x25519.X25519PublicKey
+        ],
     ) -> bytes:
+        """Calculate shared secret using ECDH.
+
+        Requires PIN verification.
+
+        :param slot: The slot.
+        :param peer_public_key: The peer's public key.
+        """
         slot = SLOT(slot)
         key_type = KEY_TYPE.from_public_key(peer_public_key)
         if key_type.algorithm != ALGORITHM.EC:
@@ -671,12 +868,21 @@ class PivSession:
         logger.debug(
             f"Performing key agreement with key in slot {slot} of type {key_type}"
         )
-        data = peer_public_key.public_bytes(
-            Encoding.X962, PublicFormat.UncompressedPoint
-        )
+        if key_type == KEY_TYPE.X25519:
+            data = peer_public_key.public_bytes(Encoding.Raw, PublicFormat.Raw)
+        else:
+            data = peer_public_key.public_bytes(
+                Encoding.X962, PublicFormat.UncompressedPoint
+            )
         return self._use_private_key(slot, key_type, data, True)
 
     def get_object(self, object_id: int) -> bytes:
+        """Get object by ID.
+
+        Requires PIN verification for protected objects.
+
+        :param object_id: The object identifier.
+        """
         logger.debug(f"Reading data from object slot {hex(object_id)}")
         if object_id == OBJECT_ID.DISCOVERY:
             expected: int = OBJECT_ID.DISCOVERY
@@ -698,6 +904,13 @@ class PivSession:
             raise BadResponseError("Malformed object data", e)
 
     def put_object(self, object_id: int, data: Optional[bytes] = None) -> None:
+        """Write data to PIV object.
+
+        Requires authentication with management key.
+
+        :param object_id: The object identifier.
+        :param data: The object data.
+        """
         self.protocol.send_apdu(
             0,
             INS_PUT_DATA,
@@ -708,34 +921,67 @@ class PivSession:
         logger.info(f"Data written to object slot {hex(object_id)}")
 
     def get_certificate(self, slot: SLOT) -> x509.Certificate:
+        """Get certificate from slot.
+
+        :param slot: The slot to get the certificate from.
+        """
         slot = SLOT(slot)
         logger.debug(f"Reading certificate in slot {slot}")
         try:
             data = Tlv.parse_dict(self.get_object(OBJECT_ID.from_slot(slot)))
-        except ValueError:
+            cert_data = data[TAG_CERTIFICATE]
+            cert_info = data[TAG_CERT_INFO][0] if TAG_CERT_INFO in data else 0
+        except (ValueError, KeyError):
             raise BadResponseError("Malformed certificate data object")
 
-        cert_info = data.get(TAG_CERT_INFO)
-        if cert_info and cert_info[0] != 0:
-            raise NotSupportedError("Compressed certificates are not supported")
+        if cert_info == 1:
+            logger.debug("Certificate is compressed, decompressing...")
+            # Compressed certificate
+            cert_data = gzip.decompress(cert_data)
+        elif cert_info != 0:
+            raise NotSupportedError("Unsupported value in CertInfo")
 
         try:
-            return x509.load_der_x509_certificate(
-                data[TAG_CERTIFICATE], default_backend()
-            )
+            return x509.load_der_x509_certificate(cert_data, default_backend())
         except Exception as e:
             raise BadResponseError("Invalid certificate", e)
 
-    def put_certificate(self, slot: SLOT, certificate: x509.Certificate) -> None:
+    def put_certificate(
+        self, slot: SLOT, certificate: x509.Certificate, compress: bool = False
+    ) -> None:
+        """Import certificate to slot.
+
+        Requires authentication with management key.
+
+        :param slot: The slot to import the certificate to.
+        :param certificate: The certificate to import.
+        :param compress: If the certificate should be compressed or not.
+        """
         slot = SLOT(slot)
         logger.debug(f"Storing certificate in slot {slot}")
         cert_data = certificate.public_bytes(Encoding.DER)
+        logger.debug(f"Certificate is {len(cert_data)} bytes, compression={compress}")
+        if compress:
+            cert_info = b"\1"
+            cert_data = gzip.compress(cert_data)
+            logger.debug(f"Compressed size: {len(cert_data)} bytes")
+        else:
+            cert_info = b"\0"
         data = (
-            Tlv(TAG_CERTIFICATE, cert_data) + Tlv(TAG_CERT_INFO, b"\0") + Tlv(TAG_LRC)
+            Tlv(TAG_CERTIFICATE, cert_data)
+            + Tlv(TAG_CERT_INFO, cert_info)
+            + Tlv(TAG_LRC)
         )
         self.put_object(OBJECT_ID.from_slot(slot), data)
+        logger.info(f"Certificate written to slot {slot}, compression={compress}")
 
     def delete_certificate(self, slot: SLOT) -> None:
+        """Delete certificate.
+
+        Requires authentication with management key.
+
+        :param slot: The slot to delete the certificate from.
+        """
         slot = SLOT(slot)
         logger.debug(f"Deleting certificate in slot {slot}")
         self.put_object(OBJECT_ID.from_slot(slot))
@@ -750,12 +996,21 @@ class PivSession:
         pin_policy: PIN_POLICY = PIN_POLICY.DEFAULT,
         touch_policy: TOUCH_POLICY = TOUCH_POLICY.DEFAULT,
     ) -> None:
+        """Import a private key to slot.
+
+        Requires authentication with management key.
+
+        :param slot: The slot to import the key to.
+        :param private_key: The private key to import.
+        :param pin_policy: The PIN policy.
+        :param touch_policy: The touch policy.
+        """
         slot = SLOT(slot)
         key_type = KEY_TYPE.from_public_key(private_key.public_key())
         check_key_support(self.version, key_type, pin_policy, touch_policy, False)
         ln = key_type.bit_len // 8
-        numbers = private_key.private_numbers()
         if key_type.algorithm == ALGORITHM.RSA:
+            numbers = private_key.private_numbers()
             numbers = cast(rsa.RSAPrivateNumbers, numbers)
             if numbers.public_numbers.e != 65537:
                 raise NotSupportedError("RSA exponent must be 65537")
@@ -767,7 +1022,15 @@ class PivSession:
                 + Tlv(0x04, int2bytes(numbers.dmq1, ln))
                 + Tlv(0x05, int2bytes(numbers.iqmp, ln))
             )
+        elif key_type in (KEY_TYPE.ED25519, KEY_TYPE.X25519):
+            data = Tlv(
+                0x07 if key_type == KEY_TYPE.ED25519 else 0x08,
+                private_key.private_bytes(
+                    Encoding.Raw, PrivateFormat.Raw, NoEncryption()
+                ),
+            )
         else:
+            numbers = private_key.private_numbers()
             numbers = cast(ec.EllipticCurvePrivateNumbers, numbers)
             data = Tlv(0x06, int2bytes(numbers.private_value, ln))
         if pin_policy:
@@ -789,6 +1052,15 @@ class PivSession:
         pin_policy: PIN_POLICY = PIN_POLICY.DEFAULT,
         touch_policy: TOUCH_POLICY = TOUCH_POLICY.DEFAULT,
     ) -> Union[rsa.RSAPublicKey, ec.EllipticCurvePublicKey]:
+        """Generate private key in slot.
+
+        Requires authentication with management key.
+
+        :param slot: The slot to generate the private key in.
+        :param key_type: The key type.
+        :param pin_policy: The PIN policy.
+        :param touch_policy: The touch policy.
+        """
         slot = SLOT(slot)
         key_type = KEY_TYPE(key_type)
         check_key_support(self.version, key_type, pin_policy, touch_policy, True)
@@ -808,19 +1080,52 @@ class PivSession:
         return _parse_device_public_key(key_type, Tlv.unpack(0x7F49, response))
 
     def attest_key(self, slot: SLOT) -> x509.Certificate:
+        """Attest key in slot.
+
+        :param slot: The slot where the key has been generated.
+        :return: A X.509 certificate.
+        """
         require_version(self.version, (4, 3, 0))
         slot = SLOT(slot)
         response = self.protocol.send_apdu(0, INS_ATTEST, slot, 0)
         logger.debug(f"Attested key in slot {slot}")
         return x509.load_der_x509_certificate(response, default_backend())
 
+    def move_key(self, from_slot: SLOT, to_slot: SLOT) -> None:
+        """Move key from one slot to another.
+
+        Requires authentication with management key.
+
+        :param from_slot: The slot containing the key to move.
+        :param to_slot: The new slot to move the key to.
+        """
+        require_version(self.version, (5, 7, 0))
+        from_slot = SLOT(from_slot)
+        to_slot = SLOT(to_slot)
+        logger.debug(f"Moving key from slot {from_slot} to {to_slot}")
+        self.protocol.send_apdu(0, INS_MOVE_KEY, to_slot, from_slot)
+        logger.info(f"Key moved from slot {from_slot} to {to_slot}")
+
+    def delete_key(self, slot: SLOT) -> None:
+        """Delete a key in a slot.
+
+        Requires authentication with management key.
+
+        :param slot: The slot containing the key to delete.
+        """
+        require_version(self.version, (5, 7, 0))
+        slot = SLOT(slot)
+        logger.debug(f"Deleting key in slot {slot}")
+        self.protocol.send_apdu(0, INS_MOVE_KEY, 0xFF, slot)
+        logger.info(f"Key deleted in slot {slot}")
+
     def _change_reference(self, ins, p2, value1, value2):
         try:
             self.protocol.send_apdu(
                 0, ins, 0, p2, _pin_bytes(value1) + _pin_bytes(value2)
             )
         except ApduError as e:
-            retries = _retries_from_sw(self.version, e.sw)
+            retries = _retries_from_sw(e.sw)
             if retries is None:
                 raise
             if p2 == PIN_P2:
